# Web utility

**Web front-end** toolkit based on [TypeScript][1]

[![NPM Dependency](https://david-dm.org/EasyWebApp/web-utility.svg)][2]
[![CI & CD](https://github.com/EasyWebApp/web-utility/actions/workflows/main.yml/badge.svg)][3]

[![Open in Visual Studio Code](https://open.vscode.dev/badges/open-in-vscode.svg)][4]

[![NPM](https://nodei.co/npm/web-utility.png?downloads=true&downloadRank=true&stars=true)][5]

## Installation

```shell
npm install web-utility
```

`index.html`

```html
<head>
    <script src="https://polyfill.app/api/polyfill?features=regenerator-runtime,url,scroll-behavior,intersection-observer"></script>
    <script src="https://cdn.jsdelivr.net/npm/fast-text-encoding@1.0.3/text.min.js"></script>
</head>
```

`tsconfig.json`

```json
{
    "compilerOptions": {
        "module": "ES2021",
        "moduleResolution": "Node",
        "downlevelIteration": true,
        "lib": ["ES2021", "DOM", "DOM.Iterable"]
    }
}
```

## Usage

[API document](https://web-cell.dev/web-utility/)

### CSS Animation

1. [Watch Swipe](https://github.com/EasyWebApp/BootCell/blob/11c5d6f/source%2FMedia%2FCarousel.tsx#L200-L218)

2. [Simple Hover](https://github.com/EasyWebApp/BootCell/blob/a41bbc1/source/Prompt/Tooltip.tsx#L38-L43)

3. [Switch with `await`](https://github.com/EasyWebApp/BootCell/blob/a41bbc1/source/Content/TabList.tsx#L77-85)

4. [Toggle with Inline Styles](https://github.com/EasyWebApp/BootCell/blob/a41bbc1/source/Content/Collapse.tsx#L19-L38)

5. [Work with Existed Classes](https://github.com/EasyWebApp/BootCell/blob/a41bbc1/source/Content/Carousel.tsx#L82-L99)

### Function Cache

```typescript
import { cache } from 'web-utility';

const getToken = cache(async (cleaner, code) => {
    const { access_token, expires_in } = await (
        await fetch(`https://example.com/access_token?code=${code}`)
    ).json();

    setTimeout(cleaner, expires_in * 1000);

    return access_token;
}, 'Get Token');

Promise.all([getToken('xxx'), getToken('yyy')]).then(([first, second]) =>
    console.assert(
        first === second,
        'Getting token for many times should return the same before deadline'
    )
);
```

### jQuery-like DOM event delegation

```javascript
import { delegate } from 'web-utility';

document.addEventListener(
    'click',
    delegate('a[href]', (event, link) => {
        event.preventDefault();

        console.log(link.href);
    })
);
```

### Message Channel

`index.ts`

```typescript
import { createMessageServer } from 'web-utility';

createMessageServer({
    preset: () => ({ test: 1 })
});
```

`iframe.ts`

```typescript
import { createMessageClient } from 'web-utility';

const request = createMessageClient(globalThis.parent);

(async () => {
    console.log(await request('preset')); // { test: 1 }
})();
```

### Service Worker updating

```javascript
import { serviceWorkerUpdate } from 'web-utility';

const { serviceWorker } = window.navigator;

serviceWorker
    ?.register('/sw.js')
    .then(serviceWorkerUpdate)
    .then(worker => {
        if (window.confirm('New version of this Web App detected, update now?'))
            // Trigger the message callback listened in the Service Worker
            // generated by Workbox CLI
            worker.postMessage({ type: 'SKIP_WAITING' });
    });

serviceWorker?.addEventListener('controllerchange', () =>
    window.location.reload()
);
```

### Internationalization

`source/i18n/en-US.ts`

```typescript
export enum en_US {
    title = 'Title',
    name = 'Name'
}
```

`source/i18n/zh-CN.ts`

```typescript
export enum zh_CN {
    title = '标题',
    name = '名称'
}
```

`source/index.tsx`

```javascript
import { render, createCell } from 'web-cell';
import { documentReady, bootI18n, textJoin } from 'web-utility';

import { en_US } from './i18n/en-US';
import { zh_CN } from './i18n/zh-CN';

console.log(
    navigator.languages.includes('zh-CN'),  // true
    document.documentElement.lang           // ''
);
const { language, words } = bootI18n({
    'en-US': en_US,
    'zh-CN': zh_CN
});
documentReady.then(() =>
    render(
        <h1>{textJoin(words.title, words.name, 'test', 'example')}</h1>
        // <h1>标题名称 test example</h1>
    );
);
```

[1]: https://www.typescriptlang.org/
[2]: https://david-dm.org/EasyWebApp/web-utility
[3]: https://github.com/EasyWebApp/web-utility/actions/workflows/main.yml
[4]: https://open.vscode.dev/EasyWebApp/web-utility
[5]: https://nodei.co/npm/web-utility/
